LIBRARY IEEE;
use ieee.std_logic_1164.ALL;
use ieee.numeric_std.ALL;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

--Library UNISIM;
--use UNISIM.vcomponents.all;

ENTITY bin2gray_p3 IS
   PORT (
      gray_reg                : OUT std_logic_vector(3 DOWNTO 0);   
      bin                     : IN std_logic_vector(3 DOWNTO 0);   
      clk                     : IN std_logic;   
      reset                   : IN std_logic);   
END bin2gray_p3;

ARCHITECTURE arch_bin2gray_p3 OF bin2gray_p3 IS

component FDP
	
	port
	(
		Q : out std_logic;
		C : in std_logic;
		D : in std_logic;
		PRE : in std_logic
	);
end component;

   SIGNAL gray         :  std_logic_vector(3 DOWNTO 0);   --combinatorial

BEGIN

   PROCESS (bin)
   BEGIN
      CASE bin IS
         WHEN "0000" => gray <= "0010";    
         WHEN "0001" => gray <= "0110";    
         WHEN "0010" => gray <= "0111";    
         WHEN "0011" => gray <= "0101";    
         WHEN "0100" => gray <= "0100";    
         WHEN "0101" => gray <= "1100";    
         WHEN "0110" => gray <= "1101";    
         WHEN "0111" => gray <= "1111";    
         WHEN "1000" => gray <= "1110";    
         WHEN "1001" => gray <= "1010";    
         WHEN "1010" => gray <= "1011";    
         WHEN "1011" => gray <= "1001";    
         WHEN "1100" => gray <= "1000";    
         WHEN "1101" => gray <= "0000";    
         WHEN "1110" => gray <= "0001";    
         WHEN "1111" => gray <= "0011";    
         WHEN OTHERS => NULL;
         
      END CASE;
   END PROCESS;
   
   -- these registers are set to 1's rather than 0's on reset to
   -- prevent an erroneous ALMOSTFULL status for 1 WRCLK period
   -- after the trailing edge of RST.

   r0 : FDP 
      PORT MAP (
         Q => gray_reg(0),
         C => clk,
         PRE => reset,
         D => gray(0));   
   

   r1 : FDP 
      PORT MAP (
         Q => gray_reg(1),
         C => clk,
         PRE => reset,
         D => gray(1));   
   

   r2 : FDP 
      PORT MAP (
         Q => gray_reg(2),
         C => clk,
         PRE => reset,
         D => gray(2));   
   

   r3 : FDP 
      PORT MAP (
         Q => gray_reg(3),
         C => clk,
         PRE => reset,
         D => gray(3));   
   

END arch_bin2gray_p3;
