--************************************************************************
-- ************************************************************************
-- *
-- *    File Name:  fifo_sol3.vhd
-- *         Date:  05/08/06
-- *       Author:  John Logue
-- *        Email:  john.logue@xilinx.com
-- *      Company:  Xilinx
-- *
-- *                Copyright (c) 2005 Xilinx, Inc.
-- *                All rights reserved
-- *
-- ************************************************************************
-- (Solution number 3).
-- Below are the Gray-converted codes for the address MSBs
-- ----------------------------------------------------
-- | Binary    Gray | Binary+1  Gray | Binary-1  Gray |
-- | ------    ---- | --------  ---- | --------  ---- |
-- | 0000      0000 | 0001      0001 | 1111      1000 |
-- | 0001      0001 | 0010      0011 | 0000      0000 |
-- | 0010      0011 | 0011      0010 | 0001      0001 |
-- | 0011      0010 | 0100      0110 | 0010      0011 |
-- | 0100      0110 | 0101      0111 | 0011      0010 |
-- | 0101      0111 | 0110      0101 | 0100      0110 |
-- | 0110      0101 | 0111      0100 | 0101      0111 |
-- | 0111      0100 | 1000      1100 | 0110      0101 |
-- | 1000      1100 | 1001      1101 | 0111      0100 |
-- | 1001      1101 | 1010      1111 | 1000      1100 |
-- | 1010      1111 | 1011      1110 | 1001      1101 |
-- | 1011      1110 | 1100      1010 | 1010      1111 |
-- | 1100      1010 | 1101      1011 | 1011      1110 |
-- | 1101      1011 | 1110      1001 | 1100      1010 |
-- | 1110      1001 | 1111      1000 | 1101      1011 |
-- | 1111      1000 | 0000      0000 | 1110      1001 |
-- ----------------------------------------------------
LIBRARY IEEE;
use ieee.std_logic_1164.ALL;
use ieee.numeric_std.ALL;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

--Library UNISIM;
--use UNISIM.vcomponents.all;

ENTITY fifo_sol3 IS
   PORT (
      ALMOSTEMPTY             : OUT std_logic;   
      ALMOSTFULL              : OUT std_logic;   
      DO                      : OUT std_logic_vector(31 DOWNTO 0);   
      DOP                     : OUT std_logic_vector(3 DOWNTO 0);   
      EMPTY                   : OUT std_logic;   
      FULL                    : OUT std_logic;   
      RDCOUNT                 : OUT std_logic_vector(11 DOWNTO 0);   
      RDERR                   : OUT std_logic;   
      WRCOUNT                 : OUT std_logic_vector(11 DOWNTO 0);   
      WRERR                   : OUT std_logic;   
      DI                      : IN std_logic_vector(31 DOWNTO 0);   
      DIP                     : IN std_logic_vector(3 DOWNTO 0);   
      RDCLK                   : IN std_logic;   
      RDEN                    : IN std_logic;   
      RST                     : IN std_logic;   
      WRCLK                   : IN std_logic;   
      WREN                    : IN std_logic);   
END fifo_sol3;

ARCHITECTURE translated OF fifo_sol3 IS

component bin2gray 
   PORT (
      gray_reg                : OUT std_logic_vector(3 DOWNTO 0);   
      bin                     : IN std_logic_vector(3 DOWNTO 0);   
      clk                     : IN std_logic;   
      reset                   : IN std_logic);   
end component;

component bin2gray_m1 
   PORT (
      gray_reg                : OUT std_logic_vector(3 DOWNTO 0);   
      bin                     : IN std_logic_vector(3 DOWNTO 0);   
      clk                     : IN std_logic;   
      reset                   : IN std_logic);   
end component;

component bin2gray_p1 
   PORT (
      gray_reg                : OUT std_logic_vector(3 DOWNTO 0);   
      bin                     : IN std_logic_vector(3 DOWNTO 0);   
      clk                     : IN std_logic;   
      reset                   : IN std_logic);   
end component;

component bin2gray_p2 
   PORT (
      gray_reg                : OUT std_logic_vector(3 DOWNTO 0);   
      bin                     : IN std_logic_vector(3 DOWNTO 0);   
      clk                     : IN std_logic;   
      reset                   : IN std_logic);   
end component;

component bin2gray_p3 
   PORT (
      gray_reg                : OUT std_logic_vector(3 DOWNTO 0);   
      bin                     : IN std_logic_vector(3 DOWNTO 0);   
      clk                     : IN std_logic;   
      reset                   : IN std_logic);   
end component;

component FDC
	port (
		Q : out std_logic;
		C : in std_logic;
		CLR : in std_logic;
		D : in std_logic);
end component;

component FDP
	port (
		Q : out std_logic;
		C : in std_logic;
		D : in std_logic;
		PRE : in std_logic);
end component;

component FIFO16
	generic
	(
		ALMOST_FULL_OFFSET : integer := 128;
		ALMOST_EMPTY_OFFSET : integer := 128;
		DATA_WIDTH : integer := 36;
		FIRST_WORD_FALL_THROUGH : string := "false"
	);
	port
	(
		ALMOSTEMPTY : out std_logic;
		ALMOSTFULL : out std_logic;
		DO : out std_logic_vector (31 downto 0);
		DOP : out std_logic_vector (3 downto 0);
		EMPTY : out std_logic;
		FULL : out std_logic;
		RDCOUNT : out std_logic_vector (11 downto 0);
		RDERR : out std_logic;
		WRCOUNT : out std_logic_vector (11 downto 0);
		WRERR : out std_logic;
		DI : in std_logic_vector (31 downto 0);
		DIP : in std_logic_vector (3 downto 0);
		RDCLK : in std_logic;
		RDEN : in std_logic;
		RST : in std_logic;
		WRCLK : in std_logic;
		WREN : in std_logic
	);
end component;

   CONSTANT  ALMOST_FULL_OFFSET    :  std_logic_vector(11 DOWNTO 0) := "000010000000";    --  not currently used
   CONSTANT  ALMOST_EMPTY_OFFSET   :  std_logic_vector(11 DOWNTO 0) := "000010000000";    --  not currently used
   CONSTANT  DATA_WIDTH            :  integer := 36;    --  not currently used
   CONSTANT  FIRST_WORD_FALL_THROUGH        :  string(1 TO 5) := "FALSE";    --  not currently used

   --  synthesis attribute ALMOST_EMPTY_OFFSET of utt is "1FC"
   --  synthesis attribute ALMOST_FULL_OFFSET of utt is "004"
   --  synthesis attribute FIRST_WORD_FALL_THROUGH of utt is "FALSE"
   --  synthesis attribute DATA_WIDTH of utt is "36"

   -- ***************** Signal Declarations *****************
   SIGNAL tempRDCOUNT             :  std_logic_vector(11 DOWNTO 0);   
   SIGNAL tempWRCOUNT             :  std_logic_vector(11 DOWNTO 0);   

   SIGNAL RDCOUNT4msbs             :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL WRCOUNT4msbs             :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL NextBit                  :  std_logic;   
   SIGNAL RAgray                   :  std_logic_vector(3 DOWNTO 0);   --  Read  Address MSBs     converted to gray and registered
   SIGNAL WAgray                   :  std_logic_vector(3 DOWNTO 0);   --  Write Address MSBs     converted to gray and registered
   SIGNAL WAgray_m1                :  std_logic_vector(3 DOWNTO 0);   --  Write Address MSBs - 1 converted to gray and registered
   SIGNAL WAgray_p1                :  std_logic_vector(3 DOWNTO 0);   --  Write Address MSBs + 1 converted to gray and registered
   SIGNAL WAgray_p2                :  std_logic_vector(3 DOWNTO 0);   --  Write Address MSBs + 2 converted to gray and registered
   SIGNAL WAgray_p3                :  std_logic_vector(3 DOWNTO 0);   --  Write Address MSBs + 3 converted to gray and registered
   SIGNAL af_p1a                   :  std_logic;   
   SIGNAL af_p1b                   :  std_logic;   
   SIGNAL af_p1c                   :  std_logic;   
   SIGNAL af_p2a                   :  std_logic;   
   SIGNAL af_p2b                   :  std_logic;   
   SIGNAL af_p2c                   :  std_logic;   
   SIGNAL af_p3a                   :  std_logic;   
   SIGNAL af_p3b                   :  std_logic;   
   SIGNAL af_p3c                   :  std_logic;   
   SIGNAL af_hold_d                :  std_logic;   
   SIGNAL af_d                     :  std_logic;   
   SIGNAL f_d                      :  std_logic;   
   SIGNAL Aempty0                  :  std_logic;   
   SIGNAL Aempty1                  :  std_logic;   
   SIGNAL ALMOSTEMPTY_sa           :  std_logic;   
   SIGNAL Aempty2                  :  std_logic;   
   SIGNAL Aempty3                  :  std_logic;   
   SIGNAL ALMOSTEMPTY_sb           :  std_logic;   
   SIGNAL ALMOSTEMPTY_all          :  std_logic;   
   SIGNAL AF1                      :  std_logic;   
   SIGNAL AF2                      :  std_logic;   
   SIGNAL AF3                      :  std_logic;   
   SIGNAL ALMOSTEMPTY_a            :  std_logic;   
   SIGNAL ALMOSTEMPTY_ap           :  std_logic;   
   SIGNAL ALMOSTEMPTY_b            :  std_logic;   
   SIGNAL ALMOSTEMPTY_bp           :  std_logic;   
   SIGNAL ALMOSTEMPTY_s0           :  std_logic;   
   SIGNAL ALMOSTEMPTY_s1           :  std_logic;   
   SIGNAL ALMOSTEMPTY_s2           :  std_logic;   
   SIGNAL ALMOSTEMPTY_s3           :  std_logic;   
   SIGNAL af_hold                  :  std_logic;   
   SIGNAL af_p1a_s                 :  std_logic;   
   SIGNAL af_p1b_s                 :  std_logic;   
   SIGNAL af_p2a_s                 :  std_logic;   
   SIGNAL af_p2b_s                 :  std_logic;   
   SIGNAL af_p3a_s                 :  std_logic;   
   SIGNAL af_p3b_s                 :  std_logic;   

BEGIN
--/* set this to the highest bit in WRCOUNT and RDCOUNT that is used. 
-- Set to 8 */

   RDCOUNT4msbs <= tempRDCOUNT(8 DOWNTO 5) ; -- RDCOUNT[`COUNT_MSB:`COUNT_MSB-3];
   WRCOUNT4msbs <= tempWRCOUNT(8 DOWNTO 5) ; -- WRCOUNT[`COUNT_MSB:`COUNT_MSB-3];
   NextBit <= tempWRCOUNT(4) ;
   
   RDCOUNT <= tempRDCOUNT;
   WRCOUNT <= tempWRCOUNT;
   -- ***************** Binary to Registered Gray Conversions *****************

   ra : bin2gray 
      PORT MAP (
         gray_reg => RAgray,
         bin => RDCOUNT4msbs,
         clk => RDCLK,
         reset => RST);   
   
   wa : bin2gray 
      PORT MAP (
         gray_reg => WAgray,
         bin => WRCOUNT4msbs,
         clk => WRCLK,
         reset => RST);   
   
   wam1 : bin2gray_m1 
      PORT MAP (
         gray_reg => WAgray_m1,
         bin => WRCOUNT4msbs,
         clk => WRCLK,
         reset => RST);   
   
   wap1 : bin2gray_p1 
      PORT MAP (
         gray_reg => WAgray_p1,
         bin => WRCOUNT4msbs,
         clk => WRCLK,
         reset => RST);   
   
   wap2 : bin2gray_p2 
      PORT MAP (
         gray_reg => WAgray_p2,
         bin => WRCOUNT4msbs,
         clk => WRCLK,
         reset => RST);   
   
   wap3 : bin2gray_p3 
      PORT MAP (
         gray_reg => WAgray_p3,
         bin => WRCOUNT4msbs,
         clk => WRCLK,
         reset => RST);   
   

   -- ***************** ALMOSTEMPTY Generation *****************

--// the equality translations are split into two sections so there is only one --//LUT in the combinatorial paths.

   --af_p1a <= CONV_STD_LOGIC(WAgray_p1(1 DOWNTO 0) = RAgray(1 DOWNTO 0)) ;

   process(WAgray_p1,RAgray)
   begin
   if (WAgray_p1(1 DOWNTO 0) = RAgray(1 DOWNTO 0)) then 
   af_p1a <= '1';
   else
   af_p1a <= '0';
   end if;
   end process;

   afp1a : FDC 
      PORT MAP (
         Q => af_p1a_s,
         C => WRCLK,
         CLR => RST,
         D => af_p1a);   
   
   --af_p1b <= CONV_STD_LOGIC(WAgray_p1(3 DOWNTO 2) = RAgray(3 DOWNTO 2)) ;
   process(WAgray_p1,RAgray)
   begin
   if (WAgray_p1(3 DOWNTO 2) = RAgray(3 DOWNTO 2)) then 
   af_p1b <= '1';
   else
   af_p1b <= '0';
   end if;
   end process;


   afp1b : FDC 
      PORT MAP (
         Q => af_p1b_s,
         C => WRCLK,
         CLR => RST,
         D => af_p1b);   
   
   af_p1c <= af_p1a_s AND af_p1b_s ;
   aful1 : FDC 
      PORT MAP (
         Q => AF1,
         C => WRCLK,
         CLR => RST,
         D => af_p1c);   
   
   --af_p2a <= CONV_STD_LOGIC(WAgray_p2(1 DOWNTO 0) = RAgray(1 DOWNTO 0))  ;

   process(WAgray_p2,RAgray)
   begin
   if (WAgray_p2(1 DOWNTO 0) = RAgray(1 DOWNTO 0)) then 
   af_p2a <= '1';
   else
   af_p2a <= '0';
   end if;
   end process;


   afp2a : FDC 
      PORT MAP (
         Q => af_p2a_s,
         C => WRCLK,
         CLR => RST,
         D => af_p2a);   
   
   --af_p2b <= CONV_STD_LOGIC(WAgray_p2(3 DOWNTO 2) = RAgray(3 DOWNTO 2)) ;

   process(WAgray_p2,RAgray)
   begin
   if (WAgray_p2(3 DOWNTO 2) = RAgray(3 DOWNTO 2)) then 
   af_p2b <= '1';
   else
   af_p2b <= '0';
   end if;
   end process;


   afp2b : FDC 
      PORT MAP (
         Q => af_p2b_s,
         C => WRCLK,
         CLR => RST,
         D => af_p2b);   
   
   af_p2c <= af_p2a_s AND af_p2b_s ;
   aful2 : FDC 
      PORT MAP (
         Q => AF2,
         C => WRCLK,
         CLR => RST,
         D => af_p2c);   
   
   --af_p3a <= CONV_STD_LOGIC(WAgray_p3(1 DOWNTO 0) = RAgray(1 DOWNTO 0))  ;

   process(WAgray_p3,RAgray)
   begin
   if (WAgray_p3(1 DOWNTO 0) = RAgray(1 DOWNTO 0)) then 
   af_p3a <= '1';
   else
   af_p3a <= '0';
   end if;
   end process;


   afp3a : FDC 
      PORT MAP (
         Q => af_p3a_s,
         C => WRCLK,
         CLR => RST,
         D => af_p3a);   
   
   --af_p3b <= CONV_STD_LOGIC(WAgray_p3(3 DOWNTO 2) = RAgray(3 DOWNTO 2)) ;

  process(WAgray_p3,RAgray)
   begin
   if (WAgray_p3(3 DOWNTO 2) = RAgray(3 DOWNTO 2)) then 
   af_p3b <= '1';
   else
   af_p3b <= '0';
   end if;
   end process;

   afp3b : FDC 
      PORT MAP (
         Q => af_p3b_s,
         C => WRCLK,
         CLR => RST,
         D => af_p3b);   
   
   af_p3c <= af_p3a_s AND af_p3b_s ;
   aful3 : FDC 
      PORT MAP (
         Q => AF3,
         C => WRCLK,
         CLR => RST,
         D => af_p3c);   
   
   af_hold_d <= AF1 OR AF2 OR AF3 ;
   afhld : FDC 
      PORT MAP (
         Q => af_hold,
         C => WRCLK,
         CLR => RST,
         D => af_hold_d);   
   
   af_d <= AF1 OR AF2 OR AF3 OR af_hold ;
   af : FDC 
      PORT MAP (
         Q => ALMOSTFULL,
         C => WRCLK,
         CLR => RST,
         D => af_d);   
   
--// ***************** FULL Generation *****************

   f_d <= af_d AND NextBit ;
   ful : FDC 
      PORT MAP (
         Q => FULL,
         C => WRCLK,
         CLR => RST,
         D => f_d);   
   
--// ***************** ALMOSTEMPTY Generation *****************

--// the equality translation is split into two sections so there is only one 
--// LUT in the combinatorial paths.

   --Aempty0 <= CONV_STD_LOGIC(WAgray(1 DOWNTO 0) = RAgray(1 DOWNTO 0)) ;

   process(WAgray,RAgray)
   begin
   if (WAgray(1 DOWNTO 0) = RAgray(1 DOWNTO 0)) then 
   Aempty0 <= '1';
   else
   Aempty0 <= '0';
   end if;
   end process;


   aemp0 : FDP 
      PORT MAP (
         Q => ALMOSTEMPTY_s0,
         C => RDCLK,
         PRE => RST,
         D => Aempty0);   
   
   --Aempty1 <= CONV_STD_LOGIC(WAgray(3 DOWNTO 2) = RAgray(3 DOWNTO 2)) ;

   process(WAgray,RAgray)
   begin
   if (WAgray(3 DOWNTO 2) = RAgray(3 DOWNTO 2)) then 
   Aempty1 <= '1';
   else
   Aempty1 <= '0';
   end if;
   end process;


   aemp1 : FDP 
      PORT MAP (
         Q => ALMOSTEMPTY_s1,
         C => RDCLK,
         PRE => RST,
         D => Aempty1);   
   
   ALMOSTEMPTY_sa <= ALMOSTEMPTY_s0 AND ALMOSTEMPTY_s1 ;
   aempa : FDP 
      PORT MAP (
         Q => ALMOSTEMPTY_a,
         C => RDCLK,
         PRE => RST,
         D => ALMOSTEMPTY_sa);   
   
   aempap : FDP 
      PORT MAP (
         Q => ALMOSTEMPTY_ap,
         C => RDCLK,
         PRE => RST,
         D => ALMOSTEMPTY_a);   
   
   --Aempty2 <= CONV_STD_LOGIC(WAgray_m1(1 DOWNTO 0) = RAgray(1 DOWNTO 0)) ;

   process(WAgray_m1,RAgray)
   begin
   if (WAgray_m1(1 DOWNTO 0) = RAgray(1 DOWNTO 0)) then 
   Aempty2 <= '1';
   else
   Aempty2 <= '0';
   end if;
   end process;


   aemp2 : FDP 
      PORT MAP (
         Q => ALMOSTEMPTY_s2,
         C => RDCLK,
         PRE => RST,
         D => Aempty2);   
   
  -- Aempty3 <= CONV_STD_LOGIC(WAgray_m1(3 DOWNTO 2) = RAgray(3 DOWNTO 2)) ;

   process(WAgray_m1,RAgray)
   begin
   if (WAgray_m1(3 DOWNTO 2) = RAgray(3 DOWNTO 2)) then 
   Aempty3 <= '1';
   else
   Aempty3 <= '0';
   end if;
   end process;


   aemp3 : FDP 
      PORT MAP (
         Q => ALMOSTEMPTY_s3,
         C => RDCLK,
         PRE => RST,
         D => Aempty3);   
   
   ALMOSTEMPTY_sb <= ALMOSTEMPTY_s2 AND ALMOSTEMPTY_s3 ;
   aempb : FDP 
      PORT MAP (
         Q => ALMOSTEMPTY_b,
         C => RDCLK,
         PRE => RST,
         D => ALMOSTEMPTY_sb);   
   
   aempbp : FDP 
      PORT MAP (
         Q => ALMOSTEMPTY_bp,
         C => RDCLK,
         PRE => RST,
         D => ALMOSTEMPTY_b);   
   
   ALMOSTEMPTY_all <= ALMOSTEMPTY_a OR ALMOSTEMPTY_ap OR ALMOSTEMPTY_b 
   OR ALMOSTEMPTY_bp ;
   
   aempae : FDP 
      PORT MAP (
         Q => ALMOSTEMPTY,
         C => RDCLK,
         PRE => RST,
         D => ALMOSTEMPTY_all);   
   
   
   -- ***************** FIFO16 Instantiation *****************

   utt : FIFO16 
      GENERIC MAP (
         ALMOST_EMPTY_OFFSET => 508,
         ALMOST_FULL_OFFSET => 4,
         DATA_WIDTH => 36,
         FIRST_WORD_FALL_THROUGH => "FALSE")
      PORT MAP (
         ALMOSTEMPTY => open,   -- this output not used
         ALMOSTFULL => open,   -- this output not used
         DO => DO,
         DOP => DOP,
         EMPTY => EMPTY,
         FULL => open,   -- this output not used
         RDCOUNT => tempRDCOUNT,
         RDERR => RDERR,
         WRCOUNT => tempWRCOUNT,
         WRERR => WRERR,
         DI => DI,
         DIP => DIP,
         RDCLK => RDCLK,
         RDEN => RDEN,
         RST => RST,
         WRCLK => WRCLK,
         WREN => WREN);   
   

END translated;

