/*************************************************************************
*
*    File Name:  fifo_third_clk.v
*         Date:  11/08/06
*
*       Author:  John Logue
*        Email:  john.logue@xilinx.com
*      Company:  Xilinx
*
*                Copyright (c) 2006 Xilinx, Inc.
*                All rights reserved
*
*************************************************************************/

// This module is a FIFO that utilizes the third clock system described 
// in PowerPoint document third_clk_nn.ppt
 
// This module may be instantiated in place of a FIFO16 module. The only difference
// is that this module includes input signal FASTCLK.

`timescale 1 ps/1 ps

module fifo_third_clk (ALMOSTEMPTY, 
                       ALMOSTFULL, 
                       DO, 
                       DOP, 
                       EMPTY, 
                       FULL, 
                       RDCOUNT, 
                       RDERR, 
                       WRCOUNT, 
                       WRERR,
                       DI, 
                       DIP, 
                       RDCLK, 
                       RDEN, 
                       RST, 
                       WRCLK, 
                       WREN,
                       FASTCLK);

parameter ALMOST_FULL_OFFSET      = 12'h080;   // these are default values, they may be overridden
parameter ALMOST_EMPTY_OFFSET     = 12'h080;   // when this module is instantiated.


output        ALMOSTEMPTY;
output        ALMOSTFULL;
output [31:0] DO;
output [3:0]  DOP;
output        EMPTY;
output        FULL;
output [11:0] RDCOUNT;
output        RDERR;
output [11:0] WRCOUNT;
output        WRERR;

input  [31:0] DI;
input  [3:0]  DIP;
input         RDCLK;
input         RDEN;
input         RST;      // hold RST active until FASTCLK is stable
input         WRCLK;
input         WREN;
input         FASTCLK;  // 400 MHz clock, driven by a BUFG

// ************************** Instantiate FIFO16 *************************


      FIFO16 # (
		.ALMOST_FULL_OFFSET(12'h001), // do not change this line
		.ALMOST_EMPTY_OFFSET(12'h1FE), // set this to FIFO16 depth - 2
		.DATA_WIDTH(36),
		.FIRST_WORD_FALL_THROUGH("FALSE")
		) fifo16_inst1 (.ALMOSTEMPTY(),
                     .ALMOSTFULL(),
                     .DO(DO),
                     .DOP(DOP),
                     .DI(DI),
                     .DIP(DIP),
                     .EMPTY(EMPTY),
                     .FULL(FULL),
                     .RDCOUNT(RDCOUNT),
                     .RDERR(RDERR),
                     .WRCOUNT(WRCOUNT),
                     .WRERR(WRERR),
                     .RDCLK(RDCLK),
                     .RDEN(RDEN),
                     .RST(RST),
                     .WRCLK(WRCLK),
                     .WREN(WREN));

// ************************** Instantiate fifo_third_clk_flags module **************************

defparam ftcf.ALMOST_EMPTY_OFFSET = ALMOST_EMPTY_OFFSET;
defparam ftcf.ALMOST_FULL_OFFSET  = ALMOST_FULL_OFFSET;

fifo_third_clk_flags ftcf (.ALMOSTEMPTY(ALMOSTEMPTY),
                           .ALMOSTFULL(ALMOSTFULL),
                           .WREN(WREN),
                           .RDEN(RDEN),
                           .RDCLK(RDCLK),
                           .WRCLK(WRCLK),
                           .FASTCLK(FASTCLK),
                           .RST(RST));

endmodule 

